# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the classes for single-qubit gates.
"""

from typing import Iterable, Optional, Union

import torch

from ...core.utils.matrix import (_h, _p, _rx, _ry, _rz, _s, _sdg, _t, _tdg,
                                  _u3, _x, _y, _z)
from .base import Gate, ParamGate


class H(Gate):
    r"""A collection of single-qubit Hadamard gates.

    The matrix form of such a gate is:

    .. math::

        H = \frac{1}{\sqrt{2}}
            \begin{bmatrix}
                1&1\\
                1&-1
            \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.

    """
    __matrix = _h(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None
    ):
        gate_info = {
            "name": "h",
            "tex": r'H',
            "api": "h", 
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, [2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return H.__matrix.to(self.device, dtype=self.dtype)


class S(Gate):
    r"""A collection of single-qubit S gates.

    The matrix form of such a gate is:

    .. math::

        S =
            \begin{bmatrix}
                1&0\\
                0&i
            \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.

    """
    __matrix = _s(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "s",
            "tex": r'S',
            "api": "s",
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, [2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return S.__matrix.to(self.device, dtype=self.dtype)


class Sdg(Gate):
    r"""A collection of single-qubit S dagger (S inverse) gates.

    The matrix form of such a gate is:

    .. math::

        S^\dagger =
            \begin{bmatrix}
                1&0\\
                0&-i
            \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.

    """
    __matrix = _sdg(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "sdg",
            "tex": r'S^\dagger',
            "api": "sdg",
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, [2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return Sdg.__matrix.to(self.device, dtype=self.dtype)


class T(Gate):
    r"""A collection of single-qubit T gates.

    The matrix form of such a gate is:

    .. math::

        T =
            \begin{bmatrix}
                1&0\\
                0&e^\frac{i\pi}{4}
            \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.

    """
    __matrix = _t(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "t",
            "tex": r'T',
            "api": "t",
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, [2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return T.__matrix.to(self.device, dtype=self.dtype)


class Tdg(Gate):
    r"""A collection of single-qubit T dagger (T inverse) gates.

    The matrix form of such a gate is:

    .. math::

        T^\dagger =
            \begin{bmatrix}
                1&0\\
                0&e^{-\frac{i\pi}{4}}
            \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.

    """
    __matrix = _tdg(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "tdg",
            "tex": r'T^\dagger',
            "api": "tdg",
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, [2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return Tdg.__matrix.to(self.device, dtype=self.dtype)


class X(Gate):
    r"""A collection of single-qubit X gates.

    The matrix form of such a gate is:

    .. math::

        \begin{bmatrix}
            0 & 1 \\
            1 & 0
        \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.

    """

    __matrix = _x(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "x",
            "tex": r'X',
            "api": "x",
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, [2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return X.__matrix.to(self.device, dtype=self.dtype)


class Y(Gate):
    r"""A collection of single-qubit Y gates.

    The matrix form of such a gate is:

    .. math::

        \begin{bmatrix}
            0 & -i \\
            i & 0
        \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.

    """
    __matrix = _y(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "y",
            "tex": r'Y',
            "api": "y",
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, [2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return Y.__matrix.to(self.device, dtype=self.dtype)


class Z(Gate):
    r"""A collection of single-qubit Z gates.

    The matrix form of such a gate is:

    .. math::

        \begin{bmatrix}
            1 & 0 \\
            0 & -1
        \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.

    """
    __matrix = _z(torch.complex128)

    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
    ):
        gate_info = {
            "name": "z",
            "tex": r'Z',
            "api": "z",
            'plot_width': 0.4,
        }
        super().__init__(
            None, qubits_idx, [2], check_legality=False, gate_info=gate_info)

    @property
    def matrix(self) -> torch.Tensor:
        return Z.__matrix.to(self.device, dtype=self.dtype)


class P(ParamGate):
    r"""A collection of single-qubit P gates.

    The matrix form of such a gate is:

    .. math::

        \begin{bmatrix}
            1 & 0 \\
            0 & e^{i\theta}
        \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    """
    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "p",
            "tex": r'P',
            "api": "p",
            "param_sharing": param_sharing,
            'plot_width': 0.9,
        }
        super().__init__(
            _p, param, 1, param_sharing, qubits_idx, check_legality=False, gate_info=gate_info, support_batch=True)


class RX(ParamGate):
    r"""A collection of single-qubit rotation gates about the x-axis.

    The matrix form of such a gate is:

    .. math::

        \begin{bmatrix}
            \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} \\
            -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2}
        \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    """
    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "rx",
            "tex": r'R_{x}',
            "api": "rx",
            "param_sharing": param_sharing,
            'plot_width': 0.9,
        }
        super().__init__(
            _rx, param, 1, param_sharing, qubits_idx, check_legality=False, gate_info=gate_info, support_batch=True)


class RY(ParamGate):
    r"""A collection of single-qubit rotation gates about the y-axis.

    The matrix form of such a gate is:

    .. math::

        \begin{bmatrix}
            \cos\frac{\theta}{2} & -\sin\frac{\theta}{2} \\
            \sin\frac{\theta}{2} & \cos\frac{\theta}{2}
        \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    """
    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "ry",
            "tex": r'R_{y}',
            "api": "ry",
            "param_sharing": param_sharing,
            'plot_width': 0.9,
        }
        super().__init__(
            _ry, param, 1, param_sharing, qubits_idx, check_legality=False, gate_info=gate_info, support_batch=True)


class RZ(ParamGate):
    r"""A collection of single-qubit rotation gates about the z-axis.

    The matrix form of such a gate is:

    .. math::

        \begin{bmatrix}
            e^{-i\frac{\theta}{2}} & 0 \\
            0 & e^{i\frac{\theta}{2}}
        \end{bmatrix}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    """
    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
            param: Optional[Union[torch.Tensor, float]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "rz",
            "tex": r'R_{z}',
            "api": "rz",
            "param_sharing": param_sharing,
            'plot_width': 0.9,
        }
        super().__init__(
            _rz, param, 1, param_sharing, qubits_idx, check_legality=False, gate_info=gate_info, support_batch=True)


class U3(ParamGate):
    r"""A collection of single-qubit rotation gates.

    The matrix form of such a gate is:

    .. math::

        \begin{align}
            U3(\theta, \phi, \lambda) =
                \begin{bmatrix}
                    \cos\frac\theta2&-e^{i\lambda}\sin\frac\theta2\\
                    e^{i\phi}\sin\frac\theta2&e^{i(\phi+\lambda)}\cos\frac\theta2
                \end{bmatrix}
        \end{align}

    Args:
        qubits_idx: Indices of the qubits on which the gates are applied. Defaults to the first qubit.
        param: Parameters of the gates. Defaults to ``None``.
        param_sharing: Whether gates in the same layer share a parameter. Defaults to ``False``.

    Raises:
        ValueError: The ``param`` must be ``torch.Tensor`` or ``float``.
    """
    def __init__(
            self, qubits_idx: Optional[Union[Iterable, int, str]] = None,
            param: Optional[Union[torch.Tensor, Iterable[float]]] = None, param_sharing: Optional[bool] = False
    ):
        gate_info = {
            "name": "u3",
            "tex": r'U',
            "api": "u3",
            "param_sharing": param_sharing,
            'plot_width': 1.65,
        }

        super().__init__(
            _u3, param, 3, param_sharing, qubits_idx, check_legality=False, gate_info=gate_info, support_batch=True)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Representations of channels
"""

import torch


def _bit_flip_kraus(prob:  torch.Tensor) -> torch.Tensor:
    prob = prob.view([1])
    
    _0 = torch.zeros_like(prob)
    kraus_oper = [
        # E0
        torch.sqrt(1 - prob), _0,
        _0, torch.sqrt(1 - prob),
        # E1
        _0, torch.sqrt(prob),
        torch.sqrt(prob), _0
    ]
    kraus_oper = torch.cat(kraus_oper).view([2, 2, 2])
    return kraus_oper + 0j


def _phase_flip_kraus(prob:  torch.Tensor) -> torch.Tensor:
    prob = prob.view([1])
    _0 = torch.zeros_like(prob)
    kraus_oper = [
        # E0
        torch.sqrt(1 - prob), _0,
        _0, torch.sqrt(1 - prob),
        # E1
        torch.sqrt(prob), _0,
        _0, -torch.sqrt(prob)
    ]
    kraus_oper = torch.cat(kraus_oper).view([2, 2, 2])
    return kraus_oper + 0j


def _bit_phase_flip_kraus(prob:  torch.Tensor) -> torch.Tensor:
    prob = prob.view([1])
    _0 = torch.zeros_like(prob)
    kraus_oper = [
        # E0
        torch.sqrt(1 - prob), _0,
        _0, torch.sqrt(1 - prob),
        # E1
        _0, -1j * torch.sqrt(prob),
        1j * torch.sqrt(prob), _0,
    ]
    kraus_oper = torch.cat(kraus_oper).view([2, 2, 2])
    return kraus_oper + 0j


def _amplitude_damping_kraus(gamma:  torch.Tensor) -> torch.Tensor:
    gamma = gamma.view([1])
    _0, _1 = torch.zeros_like(gamma), torch.ones_like(gamma)
    kraus_oper = [
        # E0
        _1, _0,
        _0, torch.sqrt(1 - gamma),
        # E1
        _0, torch.sqrt(gamma),
        _0, _0,
    ]
    kraus_oper = torch.cat(kraus_oper).view([2, 2, 2])
    return kraus_oper + 0j


def _generalized_amplitude_damping_kraus(
    gamma:  torch.Tensor, prob:  torch.Tensor) -> torch.Tensor:
    gamma, prob = gamma.view([1]), prob.view([1])
    _0 = torch.zeros_like(prob)
    kraus_oper = [
        # E0
        torch.sqrt(prob), _0,
        _0, torch.sqrt(prob) * torch.sqrt(1 - gamma),
        # E1
        _0, torch.sqrt(prob) * torch.sqrt(gamma),
        _0, _0,
        # E2
        torch.sqrt(1 - prob) * torch.sqrt(1 - gamma), _0,
        _0, torch.sqrt(1 - prob),
        # E3
        _0, _0,
        torch.sqrt(1 - prob) * torch.sqrt(gamma), _0,
    ]
    kraus_oper = torch.cat(kraus_oper).view([4, 2, 2])
    return kraus_oper + 0j


def _phase_damping_kraus(gamma:  torch.Tensor) -> torch.Tensor:
    gamma = gamma.view([1])
    _0, _1=torch.zeros_like(gamma), torch.ones_like(gamma)
    kraus_oper = [
        # E0
        _1, _0,
        _0, torch.sqrt(1 - gamma),
        # E1
        _0, _0,
        _0, torch.sqrt(gamma),
    ]
    kraus_oper = torch.cat(kraus_oper).view([2, 2, 2])
    return kraus_oper + 0j


def _depolarizing_kraus(prob:  torch.Tensor) -> torch.Tensor:
    prob = prob.view([1])
    _0 = torch.zeros_like(prob)
    kraus_oper = [
        # E0
        torch.sqrt(1 - 3 * prob / 4), _0,
        _0, torch.sqrt(1 - 3 * prob / 4),
        # E1
        _0, torch.sqrt(prob / 4),
        torch.sqrt(prob / 4), _0,
        # E2
        _0, -1j * torch.sqrt(prob / 4),
        1j * torch.sqrt(prob / 4), _0,
        # E3
        torch.sqrt(prob / 4), _0,
        _0, (-1 * torch.sqrt(prob / 4)),
    ]
    kraus_oper = torch.cat(kraus_oper).view([4, 2, 2])
    return kraus_oper

    
def _pauli_kraus(prob: torch.Tensor) -> torch.Tensor:
    prob_x, prob_y, prob_z = prob.view([3, 1])
    assert (prob_sum := torch.sum(prob)) <= 1, \
        f"The sum of input probabilities should not be greater than 1: received {prob_sum.item()}"
    prob_i = (1 - prob_sum).view([1])
    _0 = torch.zeros_like(prob_i)
    
    kraus_oper = [
        # E0
        torch.sqrt(prob_i), _0,
        _0, torch.sqrt(prob_i),
        # E1
        _0, torch.sqrt(prob_x),
        torch.sqrt(prob_x), _0,
        # E2
        _0, -1j * torch.sqrt(prob_y),
        1j * torch.sqrt(prob_y), _0,
        # E3
        torch.sqrt(prob_z), _0,
        _0, (-torch.sqrt(prob_z)),
    ]
    kraus_oper = torch.cat(kraus_oper).view([4, 2, 2])
    return kraus_oper


def _reset_kraus(prob: torch.Tensor) -> torch.Tensor:
    prob_0, prob_1 = prob.view([2, 1])
    assert (prob_sum := torch.sum(prob)) <= 1, \
        f"The sum of input probabilities should not be greater than 1: received {prob_sum.item()}"
    prob_i = (1 - prob_sum).view([1])
    _0 = torch.zeros_like(prob_i)
    kraus_oper = [
        # E0
        torch.sqrt(prob_0), _0,
        _0, _0,
        # E1
        _0, torch.sqrt(prob_0),
        _0, _0,
        # E2
        _0, _0,
        torch.sqrt(prob_1), _0,
        # E3
        _0, _0,
        _0, torch.sqrt(prob_1),
        # E4
        torch.sqrt(prob_i), _0,
        _0, torch.sqrt(prob_i),
    ]
    kraus_oper = torch.cat(kraus_oper).view([5, 2, 2])
    return kraus_oper + 0j


def _thermal_relaxation_kraus(
    const_t: torch.Tensor, exec_time: torch.Tensor) -> torch.Tensor:
    
    t1, t2 = const_t.view([2, 1])
    assert t2 <= t1, \
        f"The relaxation time T2 and T1 must satisfy T2 <= T1: received T2 {t2} and T1{t1}"
    
    exec_time = exec_time.view([1]) / 1000
    prob_reset = 1 - torch.exp(-exec_time / t1)
    prob_z = (1 - prob_reset) * (1 - torch.exp(-exec_time / t2) * torch.exp(exec_time / t1)) / 2
    prob_z = torch.zeros_like(exec_time) if torch.abs(prob_z) <= 0 else prob_z
    prob_i = 1 - prob_reset - prob_z
    _0 = torch.zeros_like(exec_time)
    kraus_oper = [
        # E0
        torch.sqrt(prob_i), _0,
        _0, torch.sqrt(prob_i),
        # E1
        torch.sqrt(prob_z), _0,
        _0, -torch.sqrt(prob_z),
        # E2
        torch.sqrt(prob_reset), _0,
        _0, _0,
        # E3
        _0, torch.sqrt(prob_reset),
        _0, _0,
    ]
    kraus_oper = torch.cat(kraus_oper).view([4, 2, 2])
    return kraus_oper + 0j


def _replacement_choi(sigma: torch.Tensor) -> torch.Tensor:
    return torch.kron(torch.eye(sigma.shape[-1]), sigma)




# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the state_vector backend.
"""

import math
from typing import List, Optional, Tuple, Union

import numpy as np
import torch

from ... import utils
from .. import MixedState, State


class PureState(State):
    r"""The pure state class.

    Args:
        data: tensor array in vector representation for quantum pure state(s).
        sys_dim: a list of dimensions for each system.
        system_seq: the system order of this state. Defaults to be from 1 to n.
        probability: tensor array for state distributions. Defaults to be 1.

    Note:
        The data is stored in the vector-form with shape :math:`(-1, d)`

    """
    def __init__(self, data: torch.Tensor, sys_dim: List[int], 
                 system_seq: Optional[List[int]] = None,
                 probability: Optional[List[torch.Tensor]] = None):
        dim = int(np.prod(sys_dim))
        self._prob = [] if probability is None else probability
        
        non_batch_len = 2 + len(self._prob)
        self._batch_dim = list(data.shape[:-non_batch_len])
        
        data = data.reshape([-1, dim])
        super().__init__(data, sys_dim, system_seq)
        
        self._switch_unitary_matrix = False # the flag to switch on the unitary matrix recording.
        
    def __getitem__(self, key: Union[int, slice]) -> 'PureState':
        assert self.batch_dim, \
            f"This state is not batched and hence cannot be indexed: received key {key}."
        return PureState(self.ket[key], self.system_dim, self.system_seq, 
                         [prob.clone() for prob in self._prob])
    
    def prob_select(self, outcome_idx: torch.Tensor, prob_idx: int = -1) -> 'PureState':
        num_prob = len(self._prob)
        if prob_idx > 0:
            prob_idx -= num_prob

        new_prob = []
        for idx, prob in enumerate(self._prob):
            if num_prob + prob_idx > idx:
                new_prob.append(prob.clone())
            else:
                new_prob.append(prob.index_select(dim=prob_idx, index=outcome_idx).squeeze(prob_idx))
        
        data_idx = prob_idx - 2
        data = self.ket.index_select(dim=data_idx, index=outcome_idx).squeeze(data_idx)
        return PureState(data, self.system_dim, self.system_seq, new_prob)
    
    def expand(self, batch_dim: List[int]) -> 'PureState':
        if self._prob != []:
            raise NotImplementedError(
                "Batch expansion of pure state with probabilistic computation is not supported."
            )
        
        expand_state = self.clone()
        expand_state._batch_dim = batch_dim
        expand_state._prob = []
        
        if np.prod(batch_dim) == np.prod(self._batch_dim):
            return expand_state
        
        non_batch_len = 1 + len(self._prob)
        expand_state._data = self._data.expand(batch_dim + [-1] * non_batch_len).reshape([-1, self.dim])
        return expand_state
    
    @property
    def backend(self) -> str:
        return "state_vector"
    
    @staticmethod
    def check(data: torch.Tensor, sys_dim: Union[int, List[int]], eps: Optional[float] = 1e-4) -> int:
        # formalize the input data with [d], [1, d], [d, 1], [..., d, 1]
        if data.ndim == 1:
            data = data.view([-1, 1])
        else:
            data = data.view(list(data.shape[:-2]) + [-1, 1])

        if isinstance(sys_dim, int):
            num_sys = int(math.log(data.shape[-2], sys_dim))
            expected_dimension = sys_dim ** num_sys
        else:
            num_sys = len(sys_dim)
            expected_dimension = int(np.prod(sys_dim))
        
        assert data.shape[-2] == expected_dimension, \
            f"Input data shape {data.shape} does not match the input system dimension {sys_dim}."
        
        if eps:
            assert torch.all(is_vec := utils.check._is_state_vector(data, eps)), \
                f"(Some) data is not a state vector: asserted {is_vec.tolist()}"
        
        return num_sys
        
    @property
    def ket(self) -> torch.Tensor:
        self.reset_sequence()
        return self._data.view(self.batch_dim + [self.dim, 1]).clone()

    @property
    def density_matrix(self) -> torch.Tensor:
        ket = self.ket
        return torch.matmul(ket, ket.mH)
    
    def to_mixed(self) -> MixedState:
        r"""Convert the pure state to mixed state computation.
        """
        return MixedState(self.density_matrix, self.system_dim, self.system_seq, 
                           [prob.clone() for prob in self._prob])
    
    def _trace(self, sys_idx: List[int]) -> MixedState:
        return self.to_mixed()._trace(sys_idx)
    
    def _transpose(self, sys_idx: List[int]) -> MixedState:
        return self.to_mixed()._transpose(sys_idx)

    @property
    def rank(self) -> int:
        return 1
    
    def normalize(self):
        return torch.div(self._data, torch.norm(self._data, dim=-1))

    def clone(self) -> 'PureState':
        data = self._data.view(self.batch_dim + [self.dim, 1]).clone()
        state = PureState(data, self.system_dim, self.system_seq, [prob.clone() for prob in self._prob])
        state._switch_unitary_matrix = self._switch_unitary_matrix
        return state

    def fit(self, backend: str) -> torch.Tensor:
        if backend == self.backend:
            return self.ket
        
        if backend == 'density_matrix':
            return self.density_matrix
        
        raise NotImplementedError(
            f"Unsupported state conversion from {self.backend} to {backend}.")

    @State.system_seq.setter
    def system_seq(self, target_seq: List[int]) -> None:
        if target_seq == self._system_seq:
            return
        
        perm_map = utils.linalg._perm_of_list(self._system_seq, target_seq)
        current_system_dim = [self._sys_dim[x] for x in self._system_seq]
        
        self._data = utils.linalg._permute_sv(self._data, perm_map, current_system_dim).contiguous()
        self._system_seq = target_seq
        
    def _record_unitary(self, unitary: torch.Tensor, sys_idx: List[int]) -> None:
        r"""The function that records the input unitary for computing the overall 
        unitary matrix of the circuit.
        
        Note:
            This function is for calling `Circuit.matrix` only
        """
        if self._batch_dim == [self.dim]:
            self._evolve_keep_dim(unitary, sys_idx)
            return
        
        applied_dim, dim = unitary.shape[-1], self.dim
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        data = self._data.view([dim, -1, applied_dim, dim // applied_dim])
        self._data = torch.matmul(unitary, data).view([-1, dim])
        
    def _evolve(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        if self._switch_unitary_matrix:
            self._record_unitary(unitary, sys_idx)
            return
        dim, _shape = self.dim, self._squeeze_shape
        
        if on_batch:
            self._batch_dim = self._batch_dim or list(unitary.shape[:-2])
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim = unitary.shape[-1]
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        unitary = unitary.view(evolve_axis + [applied_dim, applied_dim])
        data = self._data.view(_shape + [applied_dim, dim // applied_dim])
        self._data = torch.matmul(unitary, data).view([-1, dim])
    
    def _evolve_ctrl(self, unitary: torch.Tensor, index: int, sys_idx: List[Union[int, List[int]]], on_batch: bool = True) -> None:
        ctrl_idx = sys_idx[0]
        ctrl_dim = np.prod([self.system_dim[idx] for idx in ctrl_idx])
        
        sys_idx = ctrl_idx + sys_idx[1:]
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        if self._switch_unitary_matrix:
            proj = torch.zeros([ctrl_dim, ctrl_dim])
            proj[index, index] = 1
            unitary = torch.kron(proj, unitary) + torch.kron(torch.eye(ctrl_dim) - proj, torch.eye(unitary.shape[-1]).expand_as(unitary))
            self._record_unitary(unitary, sys_idx)
            return
        dim, _shape = self.dim, self._squeeze_shape
        
        if on_batch:
            self._batch_dim = self._batch_dim or list(unitary.shape[:-2])
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim = unitary.shape[-1]
        unitary = unitary.view(evolve_axis + [applied_dim, applied_dim])
        
        other_dim = dim // ctrl_dim
        data = self._data.squeeze().expand(self.batch_dim + [dim])
        
        data = data.view(_shape + [ctrl_dim, applied_dim, other_dim // applied_dim]).clone()
        data[:, :, index] = torch.matmul(unitary, data[:, :, index])
        
        self._data = data.view([-1, dim])
        
    def _evolve_keep_dim(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        unitary_batch_dim = list(unitary.shape[:-2])
        dim, _shape = self.dim, self._squeeze_shape
        
        if on_batch:
            self._batch_dim = (self._batch_dim or unitary_batch_dim[:-1]) + unitary_batch_dim[-1:]
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim, num_unitary = unitary.shape[-1], int(np.prod(unitary_batch_dim[-1:]))
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        unitary = unitary.view(evolve_axis + [num_unitary, applied_dim, applied_dim])
        data = self._data.view(_shape + [1, applied_dim, dim // applied_dim])
        self._data = torch.matmul(unitary, data).view([-1, dim])
    
    def _expec_val(self, obs: torch.Tensor, sys_idx: List[int]) -> torch.Tensor:
        dim = self.dim
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        applied_dim, num_obs = obs.shape[-1], obs.shape[-3]
        state = self._data.view([-1, 1, applied_dim, dim // applied_dim])
        measured_state = torch.matmul(obs, state).view([-1, num_obs, dim])

        return torch.linalg.vecdot(self._data.unsqueeze(-2), measured_state).view(self.batch_dim + [num_obs])
    
    def _expec_state(self, prob_idx: List[int]) -> 'PureState':
        dim, num_prob = self.dim, len(self._prob)
        batch_prob_len = len(self._prob[-1].shape[:-num_prob])
        prob = self._joint_probability(prob_idx)
        
        states = self._data.view([-1] + self._prob_dim + [dim])
        prob = prob.view(list(prob.shape) + [1] * (self._prob[-1].dim() - prob.dim() + 1))
        prob_state = torch.mul(prob, states)
        
        sum_idx = [idx + 1 for idx in prob_idx]
        expectation = prob_state.sum(sum_idx)

        new_prob = []
        if len(prob_idx) != num_prob:
            new_prob = [prob.clone() for idx, prob in enumerate(self._prob) if idx not in prob_idx]
            expectation = expectation.view(self._batch_dim + list(new_prob[-1].shape[batch_prob_len:]) + [dim, 1])
        else:
            expectation = expectation.view(self._batch_dim + [dim, 1])
        return PureState(expectation, self.system_dim, self.system_seq, new_prob)
    
    def _measure(self, measure_op: torch.Tensor, sys_idx: List[int]) -> Tuple[torch.Tensor, 'PureState']:
        new_state = self.clone()
        new_state._evolve_keep_dim(measure_op, sys_idx)
        
        data = new_state._data.view([-1, self.dim, 1])
        measure_prob = (data.mH @ data).real
        collapsed_data = data / torch.sqrt(measure_prob)
        collapsed_data[collapsed_data != collapsed_data] = 0
        
        measure_prob = measure_prob.view(new_state._batch_dim[:-1] + self._prob_dim + [-1])
        new_state._data = collapsed_data.view([-1, self.dim])
        new_state._batch_dim = new_state._batch_dim[:-1]
        new_state._prob.append(measure_prob)
        return measure_prob, new_state

    def _transform(self, *args) -> None:
        raise NotImplementedError(
            "The state vector backend does not internally support the channel conversion. \
                Please call the 'transform' function directly instead of '_transform'.")
    
    def transform(self, op: torch.Tensor, sys_idx: List[int] = None, repr_type: str = 'kraus') -> MixedState:
        return self.to_mixed().transform(op, sys_idx, repr_type)
    
    def product_trace(self, trace_state: 'PureState', trace_idx: List[int]) -> 'PureState':
        r"""Partial trace over this state, when this state is a product state
        """
        remain_seq, remain_system_dim, trace_system_dim = [], [], []
        for x in self._system_seq:
            if x in trace_idx:
                trace_system_dim.append(self.system_dim[x])
            else:
                remain_seq.append(x)
                remain_system_dim.append(self.system_dim[x])

        assert trace_state.system_dim == trace_system_dim, \
                f"Traced state does not match with the target trace system: received {trace_state.system_dim}, expect {trace_system_dim}."
        assert trace_state.numel() in [1, self.numel()], \
            f"# of traced state mismatch: received {trace_state.numel()}, expect 1 or {self.numel()}."

        trace_state = trace_state.ket.squeeze(-1)
        self.system_seq = trace_idx + remain_seq

        data = self._data.view(self.batch_dim + [self.dim])
        data = utils.linalg._ptrace_1(data, trace_state).unsqueeze(-1)
        # convert remaining sequence
        value_to_index = {value: index for index, value in enumerate(sorted(remain_seq))}
        remain_seq = [value_to_index[i] for i in remain_seq]
        return PureState(data, remain_system_dim, remain_seq, self._prob)
    
    def sqrt(self) -> torch.Tensor:
        return self.density_matrix
    
    def log(self) -> None:
        raise TypeError(
            "Cannot apply matrix logarithm to a pure state.")



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Built-in Gate matrices.
"""

import math
from typing import List, Union

import numpy as np
import torch

from .. import utils


def __get_complex_dtype(float_dtype: torch.dtype) -> torch.dtype:
    if float_dtype == torch.float32:
        complex_dtype = torch.complex64
    elif float_dtype == torch.float64:
        complex_dtype = torch.complex128
    else:
        raise ValueError(
            f"The dtype should be torch.float32 or torch.float64: received {float_dtype}")
    return complex_dtype


def _zero(dtype: torch.dtype = None, device: torch.device = None) -> torch.Tensor:
    return torch.tensor([0], dtype=dtype, device=device)


def _one(dtype: torch.dtype = None, device: torch.device = None) -> torch.Tensor:
    return torch.tensor([1], dtype=dtype, device=device)


def _phase(dim: int, dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    w = np.exp(2 * math.pi * 1j / dim)
    return torch.from_numpy(np.diag([w**i for i in range(dim)])).to(dtype)


def _shift(dim: int, dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    return torch.roll(torch.eye(dim), shifts=1, dims=0).to(dtype)


def _grover(oracle: torch.Tensor) -> torch.Tensor:
    complex_dtype = oracle.dtype
    dimension = oracle.shape[0]
    ket_zero = torch.eye(dimension, 1).to(complex_dtype)

    diffusion_op = (2 + 0j) * ket_zero @ ket_zero.T - torch.eye(dimension)
    reflection_op = torch.kron(_z(), torch.eye(dimension // 2)).to(complex_dtype)
    
    return oracle @ diffusion_op @ oracle.conj().T @ reflection_op


def _qft(num_qubits: int, dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    N = 2**num_qubits
    omega_N = np.exp(1j * 2 * math.pi / N)

    qft_mat = np.ones([N, N]).astype("complex128")
    for i in range(1, N):
        for j in range(1, N):
            qft_mat[i, j] = omega_N ** ((i * j) % N)

    return torch.tensor(qft_mat / math.sqrt(N)).to(dtype)


def _h(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    element = math.sqrt(2) / 2
    gate_matrix = [
        [element, element],
        [element, -element],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _s(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0],
        [0, 1j],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _sdg(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0],
        [0, -1j],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _t(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0],
        [0, math.sqrt(2) / 2 + math.sqrt(2) / 2 * 1j],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _tdg(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0],
        [0, math.sqrt(2) / 2 - math.sqrt(2) / 2 * 1j],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _eye(dim: int = 2, dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    return torch.eye(dim, dtype=dtype)


def _x(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [0, 1],
        [1, 0],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _y(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [0, -1j],
        [1j, 0],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _z(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0],
        [0, -1],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _p(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0, _1 = torch.zeros_like(theta), torch.ones_like(theta)
    gate_matrix = [
        _1, _0,
        _0, torch.cos(theta) + 1j * torch.sin(theta),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 2, 2])


def _rx(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    gate_matrix = [
        torch.cos(theta / 2), -1j * torch.sin(theta / 2),
        -1j * torch.sin(theta / 2), torch.cos(theta / 2),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 2, 2])


def _ry(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    gate_matrix = [
        torch.cos(theta / 2), -torch.sin(theta / 2),
        torch.sin(theta / 2), torch.cos(theta / 2),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 2, 2]) + 0j


def _rz(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0 = torch.zeros_like(theta)
    gate_matrix = [
        torch.exp(-1j * theta / 2), _0,
        _0, torch.exp(1j * theta / 2),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 2, 2])


def _u3(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 3, 1])
    theta, phi, lam = theta[:, 0], theta[:, 1], theta[:, 2]
    gate_matrix = [
        torch.cos(theta / 2), -torch.exp(1j * lam) * torch.sin(theta / 2),
        torch.exp(1j * phi) * torch.sin(theta / 2), torch.exp(1j * (phi + lam)) * torch.cos(theta / 2),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 2, 2])



# ------------------------------------------------- Split line -------------------------------------------------

    #Belows are multi-qubit matrices.



def _cnot(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 0, 1],
        [0, 0, 1, 0],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _cy(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 0, -1j],
        [0, 0, 1j, 0],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _cz(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, -1],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _swap(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0, 0, 0],
        [0, 0, 1, 0],
        [0, 1, 0, 0],
        [0, 0, 0, 1],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)



def _cp(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0, _1 = torch.zeros_like(theta), torch.ones_like(theta)
    gate_matrix = [
        _1, _0, _0, _0,
        _0, _1, _0, _0,
        _0, _0, _1, _0,
        _0, _0, _0, torch.cos(theta) + 1j * torch.sin(theta),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 4, 4])


def _crx(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0, _1 = torch.zeros_like(theta), torch.ones_like(theta)
    gate_matrix = [
        _1, _0, _0, _0, 
        _0, _1, _0, _0,
        _0, _0, torch.cos(theta / 2), -1j * torch.sin(theta / 2),
        _0, _0, -1j * torch.sin(theta / 2), torch.cos(theta / 2),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 4, 4])


def _cry(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0, _1 = torch.zeros_like(theta), torch.ones_like(theta)
    gate_matrix = [
        _1, _0, _0, _0,
        _0, _1, _0, _0,
        _0, _0, torch.cos(theta / 2), -torch.sin(theta / 2),
        _0, _0, torch.sin(theta / 2), torch.cos(theta / 2),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 4, 4]) + 0j


def _crz(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0, _1 = torch.zeros_like(theta), torch.ones_like(theta)
    gate_matrix = [
        _1, _0, _0, _0,
        _0, _1, _0, _0,
        _0, _0, torch.cos(theta / 2) - 1j * torch.sin(theta / 2), _0,
        _0, _0, _0, torch.cos(theta / 2) + 1j * torch.sin(theta / 2),
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 4, 4])


def _cu(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 4, 1])
    _0, _1 = torch.zeros_like(theta[:, -1]), torch.ones_like(theta[:, -1])

    entry22 = torch.cos(theta[:, 0] / 2) * (torch.cos(theta[:, 3]) + 1j * torch.sin(theta[:, 3]))
    entry23 = -torch.sin(theta[:, 0] / 2) * (torch.cos(theta[:, 2] + theta[:, 3]) + 1j * torch.sin(theta[:, 2] + theta[:, 3]))
    entry32 = torch.sin(theta[:, 0] / 2) * (torch.cos(theta[:, 1] + theta[:, 3]) + 1j * torch.sin(theta[:, 1] + theta[:, 3])) 
    entry33 = torch.cos(theta[:, 0] / 2) * (
        torch.cos(theta[:, 1] + theta[:, 2] + theta[:, 3])
        + 1j * torch.sin(theta[:, 1] + theta[:, 2] + theta[:, 3])
    )
    
    gate_matrix = [
        _1, _0, _0, _0,
        _0, _1, _0, _0,
        _0, _0, entry22, entry23,
        _0, _0, entry32, entry33,
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 4, 4])


def _rxx(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0 = torch.zeros_like(theta)
    _cos = torch.cos(theta / 2)
    _sin = -1j * torch.sin(theta / 2)
    
    gate_matrix = [
        _cos, _0, _0, _sin,
        _0, _cos, _sin, _0,
        _0, _sin, _cos, _0,
        _sin, _0, _0, _cos,
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 4, 4])


def _ryy(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0 = torch.zeros_like(theta)
    
    param1 = torch.cos(theta / 2)
    param2 = -1j * torch.sin(theta / 2)
    param3 = 1j * torch.sin(theta / 2)
    
    gate_matrix = [
        param1, _0, _0, param3,
        _0, param1, param2, _0,
        _0, param2, param1, _0,
        param3, _0, _0, param1,
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 4, 4])


def _rzz(theta: torch.Tensor) -> torch.Tensor:
    theta = theta.view([-1, 1])
    _0 = torch.zeros_like(theta)
    
    param1 = torch.cos(theta / 2) - 1j * torch.sin(theta / 2)
    param2 = torch.cos(theta / 2) + 1j * torch.sin(theta / 2)
    
    gate_matrix = [
        param1, _0, _0, _0,
        _0, param2, _0, _0,
        _0, _0, param2, _0,
        _0, _0, _0, param1,
    ]
    return torch.cat(gate_matrix, dim=-1).view([-1, 4, 4])


def _ms(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    val1 = math.sqrt(2) / 2
    val2 = 1j / math.sqrt(2)
    gate_matrix = [
        [val1, 0, 0, val2],
        [0, val1, val2, 0],
        [0, val2, val1, 0],
        [val2, 0, 0, val1],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _cswap(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _toffoli(dtype: torch.dtype = torch.complex128) -> torch.Tensor:
    gate_matrix = [
        [1, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0, 1, 0],
    ]
    return torch.tensor(gate_matrix, dtype=dtype)


def _universal2(theta: torch.Tensor) -> torch.Tensor:
    theta, complex_dtype = theta.view([15]), __get_complex_dtype(theta.dtype)
    unitary = _eye(4, complex_dtype)
    _cnot_gate = _cnot(complex_dtype)

    unitary = utils.linalg._unitary_transformation(
        unitary, _u3(theta[[0, 1, 2]]), qubit_idx=0, num_qubits=2
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _u3(theta[[3, 4, 5]]), qubit_idx=1, num_qubits=2
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _cnot_gate, qubit_idx=[1, 0], num_qubits=2
    )

    unitary = utils.linalg._unitary_transformation(
        unitary, _rz(theta[[6]]), qubit_idx=0, num_qubits=2
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _ry(theta[[7]]), qubit_idx=1, num_qubits=2
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _cnot_gate, qubit_idx=[0, 1], num_qubits=2
    )

    unitary = utils.linalg._unitary_transformation(
        unitary, _ry(theta[[8]]), qubit_idx=1, num_qubits=2
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _cnot_gate, qubit_idx=[1, 0], num_qubits=2
    )

    unitary = utils.linalg._unitary_transformation(
        unitary, _u3(theta[[9, 10, 11]]), qubit_idx=0, num_qubits=2
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _u3(theta[[12, 13, 14]]), qubit_idx=1, num_qubits=2
    )

    return unitary


def _universal3(theta: torch.Tensor) -> torch.Tensor:
    theta, complex_dtype = theta.view([81]), __get_complex_dtype(theta.dtype)
    unitary = _eye(8, complex_dtype)
    __h, __s, __cnot = _h(complex_dtype), _s(complex_dtype), _cnot(complex_dtype)

    psi = torch.reshape(theta[:60], shape=[4, 15])
    phi = torch.reshape(theta[60:], shape=[7, 3])

    def __block_u(_unitary, _theta):
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 2], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, _ry(_theta[0]), qubit_idx=1, num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[0, 1], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, _ry(_theta[1]), qubit_idx=1, num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[0, 1], num_qubits=3
        )

        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 2], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(_unitary, __h, qubit_idx=2, num_qubits=3)
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 0], num_qubits=3
        )

        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[0, 2], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 2], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, _rz(_theta[2]), qubit_idx=2, num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 2], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[0, 2], num_qubits=3
        )
        return _unitary

    def __block_v(_unitary, _theta):
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[2, 0], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 2], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[2, 1], num_qubits=3
        )

        _unitary = utils.linalg._unitary_transformation(
            _unitary, _ry(_theta[0]), qubit_idx=2, num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 2], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, _ry(_theta[1]), qubit_idx=2, num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 2], num_qubits=3
        )

        _unitary = utils.linalg._unitary_transformation(_unitary, __s, qubit_idx=2, num_qubits=3)
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[2, 0], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[0, 1], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[1, 0], num_qubits=3
        )

        _unitary = utils.linalg._unitary_transformation(_unitary, __h, qubit_idx=2, num_qubits=3)
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[0, 2], num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, _rz(_theta[2]), qubit_idx=2, num_qubits=3
        )
        _unitary = utils.linalg._unitary_transformation(
            _unitary, __cnot, qubit_idx=[0, 2], num_qubits=3
        )
        return _unitary

    unitary = utils.linalg._unitary_transformation(
        unitary, _universal2(psi[0]), qubit_idx=[0, 1], num_qubits=3
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _u3(phi[0, 0:3]), qubit_idx=2, num_qubits=3
    )
    unitary = __block_u(unitary, phi[1])

    unitary = utils.linalg._unitary_transformation(
        unitary, _universal2(psi[1]), qubit_idx=[0, 1], num_qubits=3
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _u3(phi[2, 0:3]), qubit_idx=2, num_qubits=3
    )
    unitary = __block_v(unitary, phi[3])

    unitary = utils.linalg._unitary_transformation(
        unitary, _universal2(psi[2]), qubit_idx=[0, 1], num_qubits=3
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _u3(phi[4, 0:3]), qubit_idx=2, num_qubits=3
    )
    unitary = __block_u(unitary, phi[5])

    unitary = utils.linalg._unitary_transformation(
        unitary, _universal2(psi[3]), qubit_idx=[0, 1], num_qubits=3
    )
    unitary = utils.linalg._unitary_transformation(
        unitary, _u3(phi[6, 0:3]), qubit_idx=2, num_qubits=3
    )
    return unitary


def _permutation(perm: List[int], system_dim: List[int]) -> torch.Tensor:
    num_system, dimension = len(perm), np.prod(system_dim)
    mat = torch.eye(dimension).view(2 * system_dim)
    idx = perm + list(range(num_system, 2 * num_system))
    return torch.permute(mat, idx).reshape([dimension, dimension])


def _param_generator(theta: torch.Tensor, generator: torch.Tensor) -> torch.Tensor:
    r"""Generate a unitary with the given parameters and generators.
    Such unitary is universal when generator forms a basis of the unitary group.
    """
    num_param = generator.shape[0]
    theta = theta.view([-1, num_param, 1, 1])
    hamiltonian = torch.sum(torch.mul(theta, generator), dim=-3)
    return torch.matrix_exp(1j * hamiltonian)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The library of torch functions used in QuAIRKit.
"""

# -------------------------- Conventions --------------------------
#
# In principle, functions inside `utils` submodule shall be   
#     - protected i.e. the name should start with "_" ;
#     - used for developers only ;
#     - used without calling other functions outside ``utils`` ;
#     - added with "utils.xxx" when used outside ``core`` module ;
#     - related to torch in/outputs only, no numpy in/outputs allowed .
#
# -----------------------------------------------------------------

from . import check, linalg, matrix, qinfo, representation



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from typing import Iterable, List, Tuple, Union

import numpy as np
import torch

from ..core import Hamiltonian, utils
from ..operator import CNOT, RX, RY, RZ, H
from .container import Layer

__all__ = ['TrotterLayer']


def map_hamilton_to_qubits_idx(hamiltonian: Hamiltonian, qubits_idx: List[int]) -> List[Hamiltonian]:
    r"""Implement the Hamiltonian on the computational qubits specified by qubit-idx, return the new Hamiltonian after mapping.

    Note:       
        This is an intrinsic function, user do not need to call this directly

    Args:
        hamiltonian (Hamiltonian): The original Hamiltonian to be mapped. It should have a method `decompose_with_sites()`
            that returns three components:
                - coeffs: A list of coefficients corresponding to each term in the Hamiltonian.
                - pauli_words: A list of Pauli strings (e.g., ['X', 'Y', 'Z']) representing the Pauli operators.
                - sites: A list of tuples specifying the qubit indices associated with each Pauli word.
        qubits_idx (List[int]): A list of integers representing the indices of the computational qubits 
            onto which the Hamiltonian will be mapped. The length of this list must be greater than the maximum 
            index present in the `sites` component of the Hamiltonian.

    Returns:
        List[Hamiltonian]: A list containing a single Hamiltonian object. This new Hamiltonian represents the 
            original Hamiltonian remapped to the specified qubits.

    Raises:
        AssertionError: If the number of qubits specified in `qubits_idx` is insufficient to represent all 
            the qubits in the original Hamiltonian.
    """
    coeffs, pauli_words, sites = hamiltonian.decompose_with_sites()
    assert len(qubits_idx) > np.max([item for sublist in sites for item in sublist if isinstance(item, int)]), "The number of qubits_idx must be bigger than the number of qubits in the Hamiltonian"
    new_terms = []

    for coeff, pauli_word, site in zip(coeffs, pauli_words, sites):
        if all(p == 'I' for p in pauli_word):
            continue
        filtered_pauli_word = [pauli for pauli in pauli_word if pauli != 'I']
        filtered_site = [s for pauli, s in zip(pauli_word, site) if pauli != 'I']
        new_site = tuple(qubits_idx[s] for s in filtered_site)

        pauli_str_parts = [
            f"{pauli}{site}"
            for pauli, site in zip(filtered_pauli_word, new_site)
        ]
        pauli_str = ', '.join(pauli_str_parts)
        new_terms.append([coeff, pauli_str])

    return [Hamiltonian(new_terms)]

def sort_pauli_word(pauli_word, site) -> tuple:
    r"""reordering the ``pauli_word`` by the value of ``site``, return the new pauli_word and site after sort.

    Note:
        This is an intrinsic function, user do not need to call this directly
    """
    sort_index = np.argsort(np.array(site))
    return ''.join(np.array(list(pauli_word))[sort_index].tolist()), np.array(site)[sort_index]


class TrotterLayer(Layer):
    r"""Add time-evolving circuits to a user-specified circuit.
    
    This circuit could approximate the time-evolving operator of a system given its Hamiltonian H,
    i.e., :math:`U_{\rm cir}~ e^{-iHt}`.

    Args:
        hamiltonian: Hamiltonian of the system whose time evolution is to be simulated.
        qubits_idx: Indices of the qubits on which the layer is applied.
        tau: Evolution time of each trotter block.
        num_steps: Number of trotter blocks that will be added in total, where ``num_steps * tau`` should be the total evolution time.
        order: Order of the Trotter-Suzuki decomposition.
        name: Name of the Hamiltonian. Defaults to 'H'.
    
    Raises:
        ValueError: The order of the trotter-suzuki decomposition should be either 1, 2 or 2k (k an integer)
        ValueError: The number of qubits_idx must be bigger than the number of qubits in the Hamiltonian
        ValueError: Grouping method ``grouping`` is not supported, valid key words: 'xyz', 'even_odd'
        ValueError: The method ``method`` is not supported, valid method keywords: 'suzuki', 'custom'

    """
    def __init__(self, hamiltonian: Hamiltonian, qubits_idx: List[int], 
                 tau: float, num_steps: int, order: int, name: str) -> None:
        super().__init__(qubits_idx, depth=num_steps, name=name)
        self._assert_qubits()
        self.tau = tau
        self.hamiltonian = hamiltonian
        
        assert hamiltonian.n_qubits == len(qubits_idx), \
            f"Number of qubits does not match: receive {qubits_idx}, expect of length {hamiltonian.n_qubits}"

        if order > 2 and order % 2 != 0 and type(order) != int:
            raise ValueError('The order of the trotter-suzuki decomposition should be either 1, 2 or 2k (k an integer)'
                             ', got order = %i' % order)
        if qubits_idx is None:
            grouped_hamiltonian = [hamiltonian]
        else:
            grouped_hamiltonian = map_hamilton_to_qubits_idx(hamiltonian, qubits_idx)

        for _ in range(num_steps):
            self._add_trotter_block(grouped_hamiltonian=grouped_hamiltonian, order=order)
    
    @property
    def fidelity(self) -> float:
        r"""The fidelity of the Trotterization.
        """
        t = self.tau * self._depth
        ideal_gate = torch.linalg.matrix_exp(-1j * t * self.hamiltonian.matrix)
        return utils.qinfo._gate_fidelity(ideal_gate, self.matrix).item()
        
    def get_latex_name(self, style: str = 'standard') -> str:
        depth = self._depth
        tau_rounded = np.round(self.tau, 3)
        time_rounded = np.round(depth * self.tau, 3)

        if style == "compact":
            if depth > 1:
                name = rf"$\approx \exp(-iH\tau) \times {depth}$, with $\tau = {tau_rounded}$"
            else:
                name = rf"$\approx \exp(-iH\tau)$, with $\tau = {tau_rounded}$"
        else:
            name = rf"$\approx \exp(-iHt)$, with $t = {time_rounded}$"
            if style == "detailed" and self.num_systems <= 6:
                fidelity_rounded = np.round(self.fidelity, 3)
                name += rf" and fidelity ${fidelity_rounded}$"
        return name

    def _add_trotter_block(self, grouped_hamiltonian, order: int) -> None:
        r"""add a Trotter block, i.e. :math:`e^{-iH\tau}`, use Trotter-Suzuki decomposition to expand it.

        Args:
            grouped_hamiltonian: list of Hamiltonian objects, this function uses these as the basic terms of Trotter-Suzuki expansion by default
            order: The order of Trotter-Suzuki expansion

        Note:
            About how to use grouped_hamiltonian: 
            For example, consider Trotter-Suzuki decomposition of the second order S2(t), if grouped_hamiltonian = [H_1, H_2], it will add Trotter circuit
            with (H_1, t/2)(H_2, t/2)(H_2, t/2)(H_1, t/2). Specifically, if user does not pre-grouping the Hamiltonians and put a single Hamiltonian object,
            this function will make canonical decomposition according to the order of this Hamiltonian: for second order, if put a single Hamiltonian H, 
            the circuit will be added with (H[0:-1:1], t/2)(H[-1:0:-1], t/2)

        Warning:
            This function is usually an intrinsic function, it does not check or correct the input. 
            To build time evolution circuit, function ``construct_trotter_circuit()`` is recommended
        """
        tau = self.tau
        if order == 1:
            self.__add_first_order_trotter_block(tau, grouped_hamiltonian)
        elif order == 2:
            self.__add_second_order_trotter_block(tau, grouped_hamiltonian)
        else:
            self.__add_higher_order_trotter_block(tau, grouped_hamiltonian, order)

    def __add_first_order_trotter_block(self, tau: float, grouped_hamiltonian: List[Hamiltonian], reverse=False) -> None:
        r"""Add a time evolution block of the first order Trotter-Suzuki decomposition

        Note:
            This is an intrinsic function, user do not need to call this directly
        """
        if not reverse:
            for hamiltonian in grouped_hamiltonian:
                assert isinstance(hamiltonian, Hamiltonian)
                # decompose the Hamiltonian into 3 lists
                coeffs, pauli_words, sites = hamiltonian.decompose_with_sites()
                term_index = 0
                while term_index < len(coeffs):
                    # get the sorted pauli_word and site (an array of qubit indices) according to their qubit indices
                    pauli_word, site = sort_pauli_word(pauli_words[term_index], sites[term_index])
                    self.add_n_pauli_gate(2 * tau * coeffs[term_index], pauli_word, site)
                    term_index += 1
        elif len(grouped_hamiltonian) == 1:
            coeffs, pauli_words, sites = grouped_hamiltonian[0].decompose_with_sites()
            for term_index in reversed(range(len(coeffs))):
                pauli_word, site = sort_pauli_word(pauli_words[term_index], sites[term_index])
                self.add_n_pauli_gate(2 * tau * coeffs[term_index], pauli_word, site)
        else:
            for hamiltonian in reversed(grouped_hamiltonian):
                assert isinstance(hamiltonian, Hamiltonian)
                coeffs, pauli_words, sites = hamiltonian.decompose_with_sites()
                for term_index in range(len(coeffs)):
                    pauli_word, site = sort_pauli_word(pauli_words[term_index], sites[term_index])
                    self.add_n_pauli_gate(2 * tau * coeffs[term_index], pauli_word, site)


    def __add_second_order_trotter_block(self, tau, grouped_hamiltonian) -> None:
        r"""Add a time evolution block of the second order Trotter-Suzuki decomposition
        
        Note:
            This is an intrinsic function, user do not need to call this directly
        """
        self.__add_first_order_trotter_block(tau / 2, grouped_hamiltonian)
        self.__add_first_order_trotter_block(tau / 2, grouped_hamiltonian, reverse=True)


    def __add_higher_order_trotter_block(self, tau, grouped_hamiltonian, order) -> None:
        r"""Add a time evolution block of the higher order (2k) Trotter-Suzuki decomposition 

        Note:
            This is an intrinsic function, user do not need to call this directly
        """
        assert order % 2 == 0
        p_values = self.get_suzuki_p_values(order)
        if order != 4:
            for p in p_values:
                self.__add_higher_order_trotter_block(p * tau, grouped_hamiltonian, order - 2)
        else:
            for p in p_values:
                self.__add_second_order_trotter_block(p * tau, grouped_hamiltonian)

    
    def get_suzuki_p_values(self, k: int) -> List:
        r"""Calculate the parameter p(k) in the Suzuki recurrence relationship.

        Args:
            k: Order of the Suzuki decomposition.

        Returns:
            A list of length five of form [p, p, (1 - 4 * p), p, p].
        """
        p = 1 / (4 - 4 ** (1 / (k - 1)))
        return [p, p, (1 - 4 * p), p, p]


    def add_n_pauli_gate(self, theta: Union[torch.Tensor, float], pauli_word: str, which_qubits: Iterable) -> None:
        r"""Add a rotation gate for a tensor product of Pauli operators, for example :math:`e^{-\theta/2 * X \otimes I \otimes X \otimes Y}`.

        Args:
            theta: Rotation angle.
            pauli_word: Pauli operators in a string format, e.g., ``"XXZ"``.
            which_qubits: List of the index of the qubit that each Pauli operator in the ``pauli_word`` acts on.

        Raises:
            ValueError: The ``which_qubits`` should be either ``list``, ``tuple``, or ``np.ndarray``.
        """
        if isinstance(which_qubits, (Tuple, List)):
            which_qubits = np.array(which_qubits)
        elif not isinstance(which_qubits, np.ndarray):
            raise ValueError('which_qubits should be either a list, tuple or np.ndarray')

        if not isinstance(theta, torch.Tensor):
            theta = torch.tensor(theta)
        # if it is a single-Pauli case, apply the single qubit rotation gate accordingly
        if len(which_qubits) == 1:
            if re.match(r'X', pauli_word[0], flags=re.I):
                self.append(RX(which_qubits[0], param=theta))
            elif re.match(r'Y', pauli_word[0], flags=re.I):
                self.append(RY(which_qubits[0], param=theta))
            elif re.match(r'Z', pauli_word[0], flags=re.I):
                self.append(RZ(which_qubits[0], param=theta))

        else:
            self.__add_n_pauli(which_qubits, pauli_word, theta)

    def __add_n_pauli(self, which_qubits, pauli_word, theta):
        which_qubits.sort()

        # Change the basis for qubits on which the acting operators are not 'Z'
        for qubit_index in range(len(which_qubits)):
            if re.match(r'X', pauli_word[qubit_index], flags=re.I):
                self.append(H([which_qubits[qubit_index]]))
            elif re.match(r'Y', pauli_word[qubit_index], flags=re.I):
                self.append(RX(which_qubits[qubit_index], param=torch.pi / 2))

        # Add a Z tensor n rotational gate
        for i in range(len(which_qubits) - 1):
            self.append(CNOT([which_qubits[i], which_qubits[i + 1]]))
        self.append(RZ(which_qubits[-1], param=theta))
        for i in reversed(range(len(which_qubits) - 1)):
            self.append(CNOT([which_qubits[i], which_qubits[i + 1]]))

        # Change the basis for qubits on which the acting operators are not 'Z'
        for qubit_index in range(len(which_qubits)):
            if re.match(r'X', pauli_word[qubit_index], flags=re.I):
                self.append(H([which_qubits[qubit_index]]))
            elif re.match(r'Y', pauli_word[qubit_index], flags=re.I):
                self.append(RX(which_qubits[qubit_index], param=-torch.pi / 2))



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


r"""
The source file of the LOCCNet class.
"""

import warnings
from typing import Dict, List, Optional, Tuple, Union

import torch
from torch.nn import ModuleDict

from quairkit import Circuit, State
from quairkit.database import bell_state, zero_state
from quairkit.operator import Collapse
from quairkit.qinfo import is_pvm, nkron

_LogicalIndex = Tuple[str, int]

class OneWayLOCCNet(torch.nn.Module):
    r"""Network for general one-way LOCC (Local Operations and Classical Communication) protocol.
    TO BE RELEASED.
    
    Args:
        party_info: a dictionary containing the information of the parties involved in the LOCC protocol.
        
    Note:
        Information of each party in `party_info` should be one of the following format:
        - {'Alice': 2} (Alice has 2 qubits)
        - {'Alice': {'num_systems': 2, 'system_dim': 3}} (Alice has 2 qutrits)
        - {'Alice': {'num_systems': 2, 'system_dim': [2, 3]}} (Alice has a qubit and a qutrit)
    
    """
    def __init__(self, party_info: Dict[str, Union[int, Dict[str, int]]]) -> None:
        super().__init__()
        
        self.__party_map: Dict[str, List[int]] = {}
        self.__state_map: Dict[List[int], State] = {}
        self.cir: ModuleDict = ModuleDict({})
        
        self.__register_party(party_info)
        
    def __register_party(self, party_info: Dict[str, Union[int, Dict[str, int]]]) -> None:
        r"""Check and process the input party information of LOCC. Generate corresponding quantum circuits.
        """
        party_map = {}, {}

        current_idx = 0
        for name, kwargs in party_info.items():
            idx_map = list(range(current_idx, current_idx + cir.num_qubits))
            cir = Circuit(kwargs) if isinstance(kwargs, int) else Circuit(**kwargs)

            party_map[name], self.cir[name] = idx_map, cir

        self.__party_map = party_map
        self.__system_dim = sum((cir.system_dim for cir in self.cir.values()), [])
        
    def __check_party(self, *party_name: str) -> None:
        r"""Check if the party is registered in the LOCC protocol.
        """
        print_str = "Available parties: " + ", ".join(self.__party_map.keys())

        for name in party_name:
            assert (
                name in self.__party_map
            ), f"{name} is not registered in the parties in this LOCC protocol\n{print_str}"
    
    def set_init_state(self, logical_indices: List[_LogicalIndex], 
                       state: Optional[State] = None) -> None:
        r"""Set the initial (Bell) state of the LOCC protocol.
        
        Args:
            logical_indices: a list of logical indices of where this state is an input
            state: the initial state of the LOCC protocol. Defaults to the (generalized) Bell state pair
        
        """
        self.__check_party(*logical_idx.keys())
        
        physical_indices, indexed_system_dim = [], []
        for name, logical_idx in logical_indices:
            physical_idx = self.__party_map[name][logical_idx]
            physical_indices.append(physical_idx)
            indexed_system_dim.append(self.__system_dim[physical_idx])

        settled_indices = {idx for indices in self.__state_map.keys() for idx in indices}
        assert set(physical_indices).isdisjoint(settled_indices), \
            "Some of input logical indices have been setup, check your codes"

        if state is None:
            assert len(physical_indices) == 2, \
                        f"Bell state can only be created for a pair of systems: received {len(physical_indices)} systems"
            state = bell_state(2, system_dim=indexed_system_dim)
        else:
            assert state.system_dim == indexed_system_dim, \
                        f"Input state dimension mismatch: expected {indexed_system_dim}, got {state.system_dim}"

        self.__state_map[physical_indices] = state
        
    def __prepare_init_state(self) -> None:
        r"""Prepare the physical initial state for the LOCC protocol.
        """
        assert self.__state_map, \
                "Initial states for all parties have not been set, please call `set_init_state` at least once"

        for indices, state in self.__state_map.items():
            self.__state_map[indices] = state.to_matrix()

        list_state, settled_indices = [], []
        for indices, state in self.__state_map.items():
            settled_indices.extend(indices)
            list_state.append(state)
        
        num_systems = len(self.__system_dim)
        if unsettled_indices := (set(range(num_systems)) - set(settled_indices)):
            # TODO wrong grammar for unsettled_indices
            warnings.warn(
                "Initial states of some systems in this protocol are not specified," + 
                "taken as zero states by default", UserWarning)
            unsettled_dim = [self.__system_dim[idx] for idx in unsettled_indices]
            list_state.append(zero_state(len(unsettled_dim), system_dim=unsettled_dim))
        unsettled_indices = sorted(list(unsettled_indices))

        init_state = nkron(*list_state)
        init_state._system_seq = settled_indices + unsettled_indices
        self.__physical_state: State = init_state
        
    def __prepare_circuit(self) -> None:
        r"""Prepare the physical initial state for the LOCC protocol.
        """
        measure_cir = Circuit(system_dim=self.__system_dim)
        for name, cir in self.cir.items():
            cir.set_system_seq(self.__party_map[name])
            cir.set_system_dim(self.__system_dim)
            cir.set_system_state(self.__physical_state)
        
        
    def set_locc(self, measure_indices: List[_LogicalIndex], act_party: str, 
                 measure_basis: Optional[torch.Tensor] = None) -> None:
        r"""Set a one-way LOCC protocol to the network.
        
        Args:
            measure_indices: a list of logical indices to be measured
            act_party: the party that performs the action w.r.t. the measurement result
            measure_idx: the index of the system that `measure_party` will measure, defaults to be all systems in the party.
            measure_op: the basis of the measurement. Defaults to the computational basis.
        
        Note:
            This function is the last step before protocol execution,
            as it will also prepare the physical initial state and quantum circuit for this protocol.
        
        """
        physical_measure_indices, measure_parties = [], []
        for name, idx in measure_indices:
            self.__check_party(name, act_party)
            physical_measure_indices.append(self.__party_map[name][idx])
            measure_parties.append(name)
        
        assert act_party not in measure_parties, \
            "The party that performs the action corresponding to the measurement should not be self-measured"
        
        physical_indices = self.__party_map[act_party]
        if measure_idx is None:
            measure_idx = physical_indices
        else:
            measure_idx = [physical_indices[idx] for idx in measure_idx]
        self.measure = Collapse(measure_idx, measure_basis=measure_basis)
        
        self.__prepare_circuit()
        
        
        
        self.__prepare_init_state()
        
    def __call__(self) -> State:
        return self.forward()
        
    def forward(self) -> State:
        r"""Run the one-way LOCC protocol and return the final state.
        
        Returns:
            The final state of the LOCC protocol.
        
        """
        return self.__physical_circuit(self.__physical_state)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Representations of channels
"""

from typing import List

import torch

from ..core import utils, State
from ..core.intrinsic import (_ArrayLike, _SingleParamLike, _StateLike,
                              _type_fetch, _type_transform)
from .set import pauli_group

__all__ = [
    "bit_flip_kraus",
    "phase_flip_kraus",
    "bit_phase_flip_kraus",
    "amplitude_damping_kraus",
    "generalized_amplitude_damping_kraus",
    "phase_damping_kraus",
    "depolarizing_kraus",
    "phase_damping_kraus",
    "depolarizing_kraus",
    "generalized_depolarizing_kraus",
    "pauli_kraus",
    "reset_kraus",
    "thermal_relaxation_kraus",
    "replacement_choi",
]


def bit_flip_kraus(prob: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of a bit flip channel with form

    .. math::

        E_0 = \sqrt{1-p} I,
        E_1 = \sqrt{p} X.

    Args:
        prob: probability :math:`p`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            prob = torch.tensor([0.5])
            kraus_operators = bit_flip_kraus(prob)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[0.7071+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.7071+0.j]],

                    [[0.0000+0.j, 0.7071+0.j],
                     [0.7071+0.j, 0.0000+0.j]]], dtype=torch.complex128)
    """
    type_str, prob = _type_fetch(prob), _type_transform(prob, "tensor")
    mat = utils.representation._bit_flip_kraus(prob)
    return _type_transform(mat, type_str)


def phase_flip_kraus(prob: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of a phase flip channel with form

    .. math::

        E_0 = \sqrt{1-p} I,
        E_1 = \sqrt{p} Z.

    Args:
        prob: probability :math:`p`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            prob = torch.tensor([0.1])
            kraus_operators = phase_flip_kraus(prob)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[ 0.9487+0.j,  0.0000+0.j],
                     [ 0.0000+0.j,  0.9487+0.j]],

                    [[ 0.3162+0.j,  0.0000+0.j],
                     [ 0.0000+0.j, -0.3162+0.j]]], dtype=torch.complex128)
    """
    type_str, prob = _type_fetch(prob), _type_transform(prob, "tensor")
    mat = utils.representation._phase_flip_kraus(prob)
    return _type_transform(mat, type_str)


def bit_phase_flip_kraus(prob: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of a bit-phase flip channel with form

    .. math::

        E_0 = \sqrt{1-p} I,
        E_1 = \sqrt{p} Y.

    Args:
        prob: probability :math:`p`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            prob = torch.tensor([0.1])
            kraus_operators = bit_phase_flip_kraus(prob)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[0.9487+0.0000j, 0.0000+0.0000j],
                     [0.0000+0.0000j, 0.9487+0.0000j]],

                    [[0.0000+0.0000j, 0.0000-0.3162j],
                     [0.0000+0.3162j, 0.0000+0.0000j]]], dtype=torch.complex128)
    """
    type_str, prob = _type_fetch(prob), _type_transform(prob, "tensor")
    mat = utils.representation._bit_phase_flip_kraus(prob)
    return _type_transform(mat, type_str)


def amplitude_damping_kraus(gamma: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of an amplitude damping channel with form

    .. math::

        E_0 =
        \begin{bmatrix}
            1 & 0 \\
            0 & \sqrt{1-\gamma}
        \end{bmatrix},
        E_1 =
        \begin{bmatrix}
            0 & \sqrt{\gamma} \\
            0 & 0
        \end{bmatrix}.

    Args:
        gamma: coefficient :math:`\gamma`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            gamma = torch.tensor(0.2)
            kraus_operators = amplitude_damping_kraus(gamma)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[1.0000+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.8944+0.j]],

                    [[0.0000+0.j, 0.4472+0.j],
                     [0.0000+0.j, 0.0000+0.j]]], dtype=torch.complex128)
    """
    type_str, gamma = _type_fetch(gamma), _type_transform(gamma, "tensor")
    mat = utils.representation._amplitude_damping_kraus(gamma)
    return _type_transform(mat, type_str)


def generalized_amplitude_damping_kraus(
        gamma: _SingleParamLike,
        prob: _SingleParamLike
) -> List[_ArrayLike]:
    r"""Kraus representation of a generalized amplitude damping channel with form

    .. math::

        E_0 = \sqrt{p} \begin{bmatrix} 1 & 0 \\ 0 & \sqrt{1-\gamma} \end{bmatrix},
        E_1 = \sqrt{p} \begin{bmatrix} 0 & \sqrt{\gamma} \\ 0 & 0 \end{bmatrix},\\
        E_2 = \sqrt{1-p} \begin{bmatrix} \sqrt{1-\gamma} & 0 \\ 0 & 1 \end{bmatrix},
        E_3 = \sqrt{1-p} \begin{bmatrix} 0 & 0 \\ \sqrt{\gamma} & 0 \end{bmatrix}.

    Args:
        gamma: coefficient :math:`\gamma`.
        prob: probability :math:`p`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            gamma = torch.tensor(0.2)
            prob = torch.tensor(0.1)
            kraus_operators = generalized_amplitude_damping_kraus(gamma, prob)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[0.3162+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.2828+0.j]],

                    [[0.0000+0.j, 0.1414+0.j],
                     [0.0000+0.j, 0.0000+0.j]],

                    [[0.8485+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.9487+0.j]],

                    [[0.0000+0.j, 0.0000+0.j],
                     [0.4243+0.j, 0.0000+0.j]]], dtype=torch.complex128)
    """
    type_str, gamma, prob = _type_fetch(gamma), _type_transform(gamma, "tensor"), _type_transform(prob, "tensor")
    mat = utils.representation._generalized_amplitude_damping_kraus(gamma, prob)
    return _type_transform(mat, type_str)


def phase_damping_kraus(gamma: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of a phase damping channel with form

    .. math::

        E_0 =
        \begin{bmatrix}
            1 & 0 \\
            0 & \sqrt{1-\gamma}
        \end{bmatrix},
        E_1 =
        \begin{bmatrix}
            0 & 0 \\
            0 & \sqrt{\gamma}
        \end{bmatrix}.

    Args:
        gamma: coefficient :math:`\gamma`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            gamma = torch.tensor(0.2)
            kraus_operators = phase_damping_kraus(gamma)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[1.0000+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.8944+0.j]],

                    [[0.0000+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.4472+0.j]]], dtype=torch.complex128)
    """
    type_str, gamma = _type_fetch(gamma), _type_transform(gamma, "tensor")
    mat = utils.representation._phase_damping_kraus(gamma)
    return _type_transform(mat, type_str)


def depolarizing_kraus(prob: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of a depolarizing channel with form

    .. math::

        E_0 = \sqrt{1-\frac{3p}{4}} I,
        E_1 = \sqrt{\frac{p}{4}} X,
        E_2 = \sqrt{\frac{p}{4}} Y,
        E_3 = \sqrt{\frac{p}{4}} Z.

    Args:
        prob: probability :math:`p`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            prob = torch.tensor(0.1)
            kraus_operators = depolarizing_kraus(prob)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[ 0.9618+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j,  0.9618+0.0000j]],

                    [[ 0.0000+0.0000j,  0.1581+0.0000j],
                     [ 0.1581+0.0000j,  0.0000+0.0000j]],

                    [[ 0.0000+0.0000j,  0.0000-0.1581j],
                     [ 0.0000+0.1581j,  0.0000+0.0000j]],

                    [[ 0.1581+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j, -0.1581+0.0000j]]], dtype=torch.complex128)
    """
    type_str, prob = _type_fetch(prob), _type_transform(prob, "tensor")
    mat = utils.representation._depolarizing_kraus(prob)
    return _type_transform(mat, type_str)


def generalized_depolarizing_kraus(prob: _SingleParamLike, num_qubits: int) -> List[_ArrayLike]:
    r"""Kraus representation of a generalized depolarizing channel with form

    .. math::

        E_0 = \sqrt{1-\frac{(D-1)p}{D}} I, \text{ where } D = 4^n,
        E_k = \sqrt{\frac{p}{D}} \sigma_k, \text{ for } 0 < k < D.

    Args:
        prob: probability :math:`p`.
        num_qubits: number of qubits :math:`n` of this channel.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            prob = torch.tensor(0.1)
            num_qubits = 1
            kraus_operators = generalized_depolarizing_kraus(prob, num_qubits)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[ 1.3601+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j,  1.3601+0.0000j]],

                    [[ 0.0000+0.0000j,  0.2236+0.0000j],
                     [ 0.2236+0.0000j,  0.0000+0.0000j]],

                    [[ 0.0000+0.0000j,  0.0000-0.2236j],
                     [ 0.0000+0.2236j,  0.0000+0.0000j]],

                    [[ 0.2236+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j, -0.2236+0.0000j]]])
    """
    type_str, prob = _type_fetch(prob), _type_transform(prob, "tensor")
    prob = prob.view([1])

    basis = pauli_group(num_qubits).to((prob + 0j).dtype)
    I, other_elements = basis[0], basis[1:]

    dim = 4 ** num_qubits
    mat = torch.stack([I * (torch.sqrt(1 - (dim - 1) * prob / dim) + 0j)] +
                      [ele * (torch.sqrt(prob / dim) + 0j) for ele in other_elements])

    return _type_transform(mat, type_str)


def pauli_kraus(prob: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of a pauli channel

    Args:
        prob: a list of three probabilities corresponding to X, Y, Z gate :math:`p`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            prob_list = torch.tensor([0.1, 0.2, 0.3])
            kraus_operators = pauli_kraus(prob_list)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[ 0.6325+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j,  0.6325+0.0000j]],

                    [[ 0.0000+0.0000j,  0.3162+0.0000j],
                     [ 0.3162+0.0000j,  0.0000+0.0000j]],

                    [[ 0.0000+0.0000j,  0.0000-0.4472j],
                     [ 0.0000+0.4472j,  0.0000+0.0000j]],

                    [[ 0.5477+0.0000j,  0.0000+0.0000j],
                     [ 0.0000+0.0000j, -0.5477+0.0000j]]], dtype=torch.complex128)
    """
    type_str, prob = _type_fetch(prob), _type_transform(prob, "tensor")
    mat = utils.representation._pauli_kraus(prob)
    return _type_transform(mat, type_str)


def reset_kraus(prob: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of a reset channel with form

    .. math::

        E_0 =
        \begin{bmatrix}
            \sqrt{p} & 0 \\
            0 & 0
        \end{bmatrix},
        E_1 =
        \begin{bmatrix}
            0 & \sqrt{p} \\
            0 & 0
        \end{bmatrix},\\
        E_2 =
        \begin{bmatrix}
            0 & 0 \\
            \sqrt{q} & 0
        \end{bmatrix},
        E_3 =
        \begin{bmatrix}
            0 & 0 \\
            0 & \sqrt{q}
        \end{bmatrix},\\
        E_4 = \sqrt{1-p-q} I.

    Args:
        prob: list of two probabilities of resetting to state :math:`|0\rangle` and :math:`|1\rangle`.

    Returns:
        a list of Kraus operators

    Examples:
        .. code-block:: python

            prob_list = torch.tensor([0.1, 0.2])
            kraus_operators = reset_kraus(prob_list)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[0.3162+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.0000+0.j]],

                    [[0.0000+0.j, 0.3162+0.j],
                     [0.0000+0.j, 0.0000+0.j]],

                    [[0.0000+0.j, 0.0000+0.j],
                     [0.4472+0.j, 0.0000+0.j]],

                    [[0.0000+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.4472+0.j]],

                    [[0.8367+0.j, 0.0000+0.j],
                     [0.0000+0.j, 0.8367+0.j]]], dtype=torch.complex128)
    """
    type_str, prob = _type_fetch(prob), _type_transform(prob, "tensor")
    mat = utils.representation._reset_kraus(prob)
    return _type_transform(mat, type_str)


def thermal_relaxation_kraus(const_t: _SingleParamLike, exec_time: _SingleParamLike) -> List[_ArrayLike]:
    r"""Kraus representation of a thermal relaxation channel

    Args:
        const_t: list of :math:`T_1` and :math:`T_2` relaxation time in microseconds.
        exec_time: quantum gate execution time in the process of relaxation in nanoseconds.

    Returns:
        a list of Kraus operators.

    Examples:
        .. code-block:: python

            const_t = torch.tensor([50, 30])
            exec_time = torch.tensor([100])
            kraus_operators = thermal_relaxation_kraus(const_t, exec_time)
            print(f'The Kraus operators are:\n{kraus_operators}')

        ::

            The Kraus operators are:
            tensor([[[ 0.9987+0.j,  0.0000+0.j],
                     [ 0.0000+0.j,  0.9987+0.j]],

                    [[ 0.0258+0.j,  0.0000+0.j],
                     [ 0.0000+0.j, -0.0258+0.j]],

                    [[ 0.0447+0.j,  0.0000+0.j],
                     [ 0.0000+0.j,  0.0000+0.j]],

                    [[ 0.0000+0.j,  0.0447+0.j],
                     [ 0.0000+0.j,  0.0000+0.j]]], dtype=torch.complex128)
    """
    type_str, const_t, exec_time = _type_fetch(const_t), _type_transform(const_t, "tensor"), _type_transform(exec_time, "tensor")
    mat = utils.representation._thermal_relaxation_kraus(const_t, exec_time)
    return _type_transform(mat, type_str)


def replacement_choi(sigma: _StateLike) -> _ArrayLike:
    r"""Choi representation of a replacement channel

    Args:
        sigma: output state of this channel.

    Returns:
        a Choi operator.

    Examples:
        .. code-block:: python

            sigma = torch.tensor([[0.8, 0.0], [0.0, 0.2]])
            choi_operator = replacement_choi(sigma)
            print(f'The Choi operator is :\n{choi_operator}')

        ::

            The Choi operator is :
            tensor([[0.8000+0.j, 0.0000+0.j, 0.0000+0.j, 0.0000+0.j],
                    [0.0000+0.j, 0.2000+0.j, 0.0000+0.j, 0.0000+0.j],
                    [0.0000+0.j, 0.0000+0.j, 0.8000+0.j, 0.0000+0.j],
                    [0.0000+0.j, 0.0000+0.j, 0.0000+0.j, 0.2000+0.j]], dtype=torch.complex128)
    """
    if isinstance(sigma, State):
        sigma = sigma.density_matrix
    type_str, sigma = _type_fetch(sigma), _type_transform(sigma, "tensor")
    mat = utils.representation._replacement_choi(sigma)
    return _type_transform(mat, type_str)



# !/usr/bin/env python3
# Copyright (c) 2023 Paddle Quantum Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The module of the hamiltonian class.
"""

import copy
import re
from typing import Iterable, List, Optional, Tuple

import numpy as np
import torch
from scipy import sparse

from . import get_device, get_dtype


class Hamiltonian:
    r"""Hamiltonian ``class`` in QuAIRKit.

    User can instantiate the ``class`` with a Pauli string.

    Args:
        pauli_str: A list of Hamiltonian information, e.g. ``[(1, 'Z0, Z1'), (2, 'I')]``
        compress: Determines whether the input list will be automatically merged (e.g. ``(1, 'Z0, Z1')`` and ``(2, 'Z1, Z0')``, these two items will be automatically merged).
         Defaults to ``True``.

    Returns:
        Create a Hamiltonian class

    Note:
        If ``compress=False``, the legitimacy of the input will not be checked.
    """

    def __init__(self, pauli_str: list, compress: Optional[bool] = True):
        self.__coefficients = None
        self.__terms = None
        self.__pauli_words_r = []
        self.__pauli_words = []
        self.__sites = []
        self.__nqubits = None
        # when internally updating the __pauli_str, be sure to set __update_flag to True
        self.__pauli_str = pauli_str
        self.__update_flag = True
        self.__decompose()
        if compress:
            self.__compress()

    def __getitem__(self, indices):
        if isinstance(indices, int):
            indices = [indices]
        elif isinstance(indices, slice):
            indices = list(range(self.n_terms)[indices])
        elif isinstance(indices, tuple):
            indices = list(indices)

        new_pauli_str = [[self.coefficients[index], ','.join(self.terms[index])] for index in indices]

        return Hamiltonian(new_pauli_str, compress=False)

    def __add__(self, h_2):
        new_pauli_str = self.pauli_str.copy()
        if isinstance(h_2, (float, int)):
            new_pauli_str.extend([[float(h_2), 'I']])
        else:
            new_pauli_str.extend(h_2.pauli_str)
        return Hamiltonian(new_pauli_str)

    def __mul__(self, other):
        new_pauli_str = copy.deepcopy(self.pauli_str)
        for i in range(len(new_pauli_str)):
            new_pauli_str[i][0] *= other
        return Hamiltonian(new_pauli_str, compress=False)

    def __sub__(self, other):
        return self.__add__(other.__mul__(-1))

    def __str__(self):
        str_out = ''
        for idx in range(self.n_terms):
            str_out += f'{self.coefficients[idx]} '
            for _ in range(len(self.terms[idx])):
                str_out += self.terms[idx][_]
                if _ != len(self.terms[idx]) - 1:
                    str_out += ', '
            if idx != self.n_terms - 1:
                str_out += '\n'
        return str_out

    def __repr__(self):
        return 'quairkit.Hamiltonian object: \n' + self.__str__()

    @property
    def n_terms(self) -> int:
        r"""Number of terms of the hamiltonian.
        """
        return len(self.__pauli_str)

    @property
    def pauli_str(self) -> List:
        r"""The Pauli string corresponding to the hamiltonian.
        """
        return self.__pauli_str

    @property
    def terms(self) -> Iterable[float]:
        r"""All items in hamiltonian, i.e. ``[['Z0, Z1'], ['I']]``.
        """
        if self.__update_flag:
            self.__decompose()
        return self.__terms

    @property
    def coefficients(self) -> List[float]:
        r"""The coefficient of each term in the Hamiltonian, i.e. ``[1.0, 2.0]``.
        """
        if self.__update_flag:
            self.__decompose()
        return self.__coefficients

    @property
    def pauli_words(self) -> List[str]:
        r"""The Pauli word of each term, i.e. ``['ZIZ', 'IIX']``.
        """
        if self.__update_flag:
            self.__decompose()
        return self.__pauli_words

    @property
    def pauli_words_r(self) -> List[str]:
        r"""A list of Pauli word (exclude I), i.e. ``['ZXZZ', 'Z', 'X']``.
        """
        if self.__update_flag:
            self.__decompose()
        return self.__pauli_words_r

    @property
    def pauli_words_matrix(self) -> List[torch.Tensor]:
        r"""The list of matrices with respect to simplied Pauli words.
        """

        def to_matrix(string):
            matrix = torch.tensor([1 + 0j], dtype=get_dtype())
            for i in range(len(string)):
                if string[i] == 'X':
                    matrix = torch.kron(
                        matrix,
                        torch.tensor([[0, 1], [1, 0]], dtype=get_dtype())
                    )
                elif string[i] == 'Y':
                    matrix = torch.kron(
                        matrix,
                        torch.tensor([[0j, -1j], [1j, 0j]], dtype=get_dtype())
                    )
                elif string[i] == 'Z':
                    matrix = torch.kron(
                        matrix,
                        torch.tensor([[1, 0], [0, -1]], dtype=get_dtype())
                    )
                elif string[i] == 'I':
                    matrix = torch.kron(
                        matrix,
                        torch.tensor([[1, 0], [0, 1]], dtype=get_dtype())
                    )
                else:
                    raise ValueError(f'wrong format of string {string}')
            return matrix

        return list(map(to_matrix, copy.deepcopy(self.pauli_words_r)))
    
    @property
    def matrix(self) -> torch.Tensor:
        r"""The matrix form of the Hamiltonian.
        """
        return torch.from_numpy(self.construct_h_matrix()).to(device=get_device(), dtype=get_dtype())

    @property
    def sites(self) -> List:
        r"""A list of qubits index corresponding to the hamiltonian.
        """
        if self.__update_flag:
            self.__decompose()
        return self.__sites

    @property
    def n_qubits(self) -> int:
        r"""Number of qubits.
        """
        if self.__update_flag:
            self.__decompose()
        return self.__nqubits

    def __decompose(self):
        r"""decompose the Hamiltonian into vairious forms

        Raises:
            Exception: Operators should be defined with a string composed of Pauli operators followed by qubit index on which it act, separated with ",". i.e. "Z0, X1"

        Note:
            This is an intrinsic function, user do not need to call this directly
            This is a fundamental function, it decomposes the input Pauli string into different forms and stores them into private variables.
        """
        self.__pauli_words = []
        self.__pauli_words_r = []
        self.__sites = []
        self.__terms = []
        self.__coefficients = []
        self.__nqubits = 1
        new_pauli_str = []
        for coefficient, pauli_term in self.__pauli_str:
            pauli_word_r = ''
            site = []
            single_pauli_terms = re.split(r',\s*', pauli_term.upper())
            self.__coefficients.append(float(coefficient))
            self.__terms.append(single_pauli_terms)
            for single_pauli_term in single_pauli_terms:
                match_i = re.match(r'I', single_pauli_term, flags=re.I)
                if match_i:
                    assert single_pauli_term[0].upper() == 'I', \
                            'The offset is defined with a sole letter "I", i.e. (3.0, "I")'
                    pauli_word_r += 'I'
                    site.append('')
                else:
                    match = re.match(r'([XYZ])([0-9]+)', single_pauli_term, flags=re.I)
                    if not match:
                        raise Exception(
                            'Operators should be defined with a string composed of Pauli operators followed' +
                            'by qubit index on which it act, separated with ",". i.e. "Z0, X1"')
                    pauli_word_r += match[1].upper()
                    assert int(match[2]) not in site, 'each Pauli operator should act on different qubit'

                    site.append(int(match[2]))
                    self.__nqubits = max(self.__nqubits, int(match[2]) + 1)
            self.__pauli_words_r.append(pauli_word_r)
            self.__sites.append(site)
            new_pauli_str.append([float(coefficient), pauli_term.upper()])

        for term_index in range(len(self.__pauli_str)):
            pauli_word = ['I' for _ in range(self.__nqubits)]
            site = self.__sites[term_index]
            for site_index in range(len(site)):
                if type(site[site_index]) == int:
                    pauli_word[site[site_index]] = self.__pauli_words_r[term_index][site_index]
            self.__pauli_words.append(''.join(pauli_word))
            self.__pauli_str = new_pauli_str
            self.__update_flag = False

    def __compress(self):
        r"""combine like terms

        Note:
            This is an intrinsic function, user do not need to call this directly
        """
        if self.__update_flag:
            self.__decompose()
        new_pauli_str = []
        flag_merged = [False for _ in range(self.n_terms)]
        for term_idx_1 in range(self.n_terms):
            if not flag_merged[term_idx_1]:
                for term_idx_2 in range(term_idx_1 + 1, self.n_terms):
                    if not flag_merged[term_idx_2] and self.pauli_words[term_idx_1] == self.pauli_words[term_idx_2]:
                        self.__coefficients[term_idx_1] += self.__coefficients[term_idx_2]
                        flag_merged[term_idx_2] = True
                if self.__coefficients[term_idx_1] != 0:
                    new_pauli_str.append([self.__coefficients[term_idx_1], ','.join(self.__terms[term_idx_1])])
        self.__pauli_str = new_pauli_str
        self.__update_flag = True

    def decompose_with_sites(self) -> Tuple[list]:
        r"""Decompose pauli_str into coefficients, a simplified form of Pauli strings, and the indices of qubits on which the Pauli operators act on.

        Returns:
            A tuple containing the following elements:
                - coefficients: the coefficient for each term.
                - pauli_words_r: the simplified form of the Pauli string for each item, e.g. the Pauli word of 'Z0, Z1, X3' is 'ZZX'. 
                - sites: a list of qubits index, e.g. the site of 'Z0, Z1, X3' is [0, 1, 3].
        """
        if self.__update_flag:
            self.__decompose()
        return self.coefficients, self.__pauli_words_r, self.__sites

    def decompose_pauli_words(self) -> Tuple[list]:
        r"""Decompose pauli_str into coefficients and Pauli strings.

        Returns:
            A tuple containing the following elements:
                - coefficients: the coefficient for each term.
                - the Pauli string for each item, e.g. the Pauli word of 'Z0, Z1, X3' is 'ZZIX'.
        """
        if self.__update_flag:
            self.__decompose()
        return self.coefficients, self.__pauli_words

    def construct_h_matrix(self, qubit_num: Optional[int] = None) -> np.ndarray:
        r"""Construct a matrix form of the Hamiltonian in Z-basis.

        Args:
            qubit_num: The number of qubits. Defaults to ``None``.

        Returns:
            The matrix form of the Hamiltonian in Z-basis.
        """
        coefs, pauli_words, sites = self.decompose_with_sites()
        if qubit_num is None:
            qubit_num = 1
            for site in sites:
                if type(site[0]) is int:
                    qubit_num = max(qubit_num, max(site) + 1)
        else:
            assert qubit_num >= self.n_qubits, "the input number of qubits must be no less than the number of qubits of this Hamiltonian"
        n_qubit = qubit_num
        h_matrix = np.zeros([2 ** n_qubit, 2 ** n_qubit])
        spin_ops = SpinOps(n_qubit, use_sparse=True)
        for idx in range(len(coefs)):
            op = coefs[idx] * sparse.eye(2 ** n_qubit)
            for site_idx in range(len(sites[idx])):
                if re.match(r'X', pauli_words[idx][site_idx], re.I):
                    op = op.dot(spin_ops.sigx_p[sites[idx][site_idx]])
                elif re.match(r'Y', pauli_words[idx][site_idx], re.I):
                    op = op.dot(spin_ops.sigy_p[sites[idx][site_idx]])
                elif re.match(r'Z', pauli_words[idx][site_idx], re.I):
                    op = op.dot(spin_ops.sigz_p[sites[idx][site_idx]])
            h_matrix += op
        return h_matrix


class SpinOps:
    r"""The spin operators in matrix forms, could be used to construct Hamiltonian matrix or spin observables.
    
    Args:
        size: Size of the system (number of qubits).
        use_sparse: Decide whether to use the sparse matrix to calculate. Default is ``False``.
    """

    def __init__(self, size: int, use_sparse: Optional[bool] = False):
        self.size = size
        self.id = sparse.eye(2)
        self.__sigz = sparse.bsr_matrix([[1, 0], [0, -1]])
        self.__sigy = sparse.bsr_matrix([[0, -1j], [1j, 0]])
        self.__sigx = sparse.bsr_matrix([[0, 1], [1, 0]])
        self.__sigz_p = []
        self.__sigy_p = []
        self.__sigx_p = []
        self.__sparse = use_sparse
        for i in range(self.size):
            self.__sigz_p.append(self.__direct_prod_op(spin_op=self.__sigz, spin_index=i))
            self.__sigy_p.append(self.__direct_prod_op(spin_op=self.__sigy, spin_index=i))
            self.__sigx_p.append(self.__direct_prod_op(spin_op=self.__sigx, spin_index=i))

    @property
    def sigz_p(self) -> List:
        r""" A list of :math:`S^z_i` operators, different elements correspond to different indices :math:`i`.
        """
        return self.__sigz_p

    @property
    def sigy_p(self) -> List:
        r""" A list of :math:`S^y_i` operators, different elements correspond to different indices :math:`i`.
        """
        return self.__sigy_p

    @property
    def sigx_p(self) -> List:
        r""" A list of :math:`S^x_i` operators, different elements correspond to different indices :math:`i`.
        """
        return self.__sigx_p

    def __direct_prod_op(self, spin_op, spin_index):
        r"""get spin operators on n-th spin (qubit) with direct product

        Args:
            spin_op: single body spin operator
            spin_index: on which spin (qubit)

        Returns:
            scipy.sparse or np.ndarray: spin operator with direct product form. The data type is specified by self.__use_sparse
        """
        s_p = copy.copy(spin_op)
        for i in range(self.size):
            if i < spin_index:
                s_p = sparse.kron(self.id, s_p)
            elif i > spin_index:
                s_p = sparse.kron(s_p, self.id)
        if self.__sparse:
            return s_p
        else:
            return s_p.toarray()


def _pauli_str_to_matrix(pauli_str: list, n: int) -> torch.Tensor:
    r"""Convert the input list form of an observable to its matrix form.

    For example, if the input ``pauli_str`` is ``[[0.7, 'z0,x1'], [0.2, 'z1']]`` and ``n=3``,
    then this function returns the observable :math:`0.7Z\otimes X\otimes I+0.2I\otimes Z\otimes I`
    in matrix form.

    Args:
        pauli_str: A list form of an observable.
        n: Number of qubits.

    Raises:
        ValueError: Only Pauli operator "I" can be accepted without specifying its position.

    Returns:
        The matrix form of the input observable.
    """
    pauli_dict = {'i': np.eye(2) + 0j, 'x': np.array([[0, 1], [1, 0]]) + 0j,
                  'y': np.array([[0, -1j], [1j, 0]]), 'z': np.array([[1, 0], [0, -1]]) + 0j}

    # Parse pauli_str; 'x0,z1,y4' to 'xziiy'
    new_pauli_str = []
    for coeff, op_str in pauli_str:
        init = list('i' * n)
        op_list = re.split(r',\s*', op_str.lower())
        for op in op_list:
            if len(op) > 1:
                pos = int(op[1:])
                assert pos < n, 'n is too small'
                init[pos] = op[0]
            elif op.lower() != 'i':
                raise ValueError('Only Pauli operator "I" can be accepted without specifying its position')
        new_pauli_str.append([coeff, ''.join(init)])

    # Convert new_pauli_str to matrix; 'xziiy' to NKron(x, z, i, i, y)
    matrices = []
    for coeff, op_str in new_pauli_str:
        sub_matrices = [pauli_dict[op.lower()] for op in op_str]
        if len(op_str) == 1:
            matrices.append(coeff * sub_matrices[0])
        else:
            mat = sub_matrices[0]
            for idx in range(1, len(sub_matrices)):
                mat = np.kron(mat, sub_matrices[idx])
            matrices.append(coeff * mat)

    return torch.tensor(sum(matrices), dtype=get_dtype())



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the basic class for the quantum gates.
"""

from typing import Any, Callable, Iterable, List, Union

import matplotlib
import torch

from ...core import OperatorInfoType, State, get_float_dtype, intrinsic, utils
from ..channel import Channel
from .visual import _base_gate_display, _base_param_gate_display


class Gate(Channel):
    r"""Base class for quantum gates.

    Args:
        matrix: the matrix of this gate. Defaults to ``None`` i.e. not specified.
        system_idx: indices of the systems that this gate acts on. Defaults to ``None``. i.e. list(range(# of acted systems)).
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
        check_legality: whether check the completeness of the matrix if provided. Defaults to ``True``.
        gate_info: information of this gate that will be placed into the gate history or plotted by a Circuit. 
            Defaults to ``None``.
    """

    def __init__(
            self, matrix: torch.Tensor = None, system_idx: Union[Iterable[Iterable[int]], Iterable[int], int] = None,
            acted_system_dim: Union[List[int], int] = 2, check_legality: bool = True, gate_info: dict = None
    ):
        assert not (matrix is None and isinstance(acted_system_dim, int)), (
            "Received None for matrix and integer for acted_system_dim: "
            "either one of them must be specified to initialize a Gate instance.")
        
        super().__init__('gate', matrix, system_idx, acted_system_dim, check_legality, gate_info)
        
        self.__matrix = matrix if matrix is None else matrix.to(dtype=self.dtype, device=self.device)
    
    def __call__(self, state: State) -> State:
        return self.forward(state)

    @property
    def matrix(self) -> torch.Tensor:
        r"""Unitary matrix of this gate

        Raises:
            ValueError: Need to specify the matrix form in this Gate instance.

        """
        if self.__matrix is None:
            raise ValueError(
                "Need to specify the matrix form in this Gate instance.")
        return self.__matrix

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float) -> float:
        r'''The display function called by circuit instance when plotting.

        Args:
            ax: the ``matplotlib.axes.Axes`` instance
            x: the start horizontal position

        Returns:
            the total width occupied

        Note:
            Users could overload this function for custom display.
        '''
        return _base_gate_display(self, ax, x)

    def _single_qubit_combine_with_threshold6(self, state: State, matrices: List[torch.Tensor]) -> None:
        r"""
        Combines single-qubit gates in a circuit with 6-qubit threshold and updates the state accordingly.

        Args:
            state: The current state, represented as a `State` object.
            matrices: A list of `torch.Tensor` objects representing the single-qubit gates to be combined.

        Returns:
            The updated state after the single-qubit gates have been combined, represented as a `State` object.
        """
        threshold_dim = 2 ** 6
        threshold_systems = threshold_dim // self.dim
        tensor_times, tensor_left = divmod(len(self.system_idx), threshold_systems)
        for threshold_idx in range(tensor_times):
            idx_beg = threshold_idx * threshold_systems
            idx_end = (threshold_idx + 1) * threshold_systems
            
            matrix = utils.linalg._nkron(*matrices[idx_beg:idx_end])
            state._evolve(matrix, sum(self.system_idx[idx_beg:idx_end], []))
        
        if tensor_left > 0:
            idx_beg = tensor_times * threshold_systems
            
            matrix = utils.linalg._nkron(*matrices[idx_beg:]) if tensor_left > 1 else matrices[idx_beg]
            state._evolve(matrix, sum(self.system_idx[idx_beg:], []))
        
        return state

    def forward(self, state: State) -> State:
        if self.num_acted_system == 1 and self.dim == 2:
            state = self._single_qubit_combine_with_threshold6(
                state=state, matrices=[self.matrix for _ in self.system_idx])
        else:
            for system_idx in self.system_idx:
                state._evolve(self.matrix, system_idx)

        return state


class ParamGate(Gate):
    r"""Base class for quantum parameterized gates.

    Args:
        generator: function that generates the unitary matrix of this gate.
        param: input parameters of quantum parameterized gates. Defaults to ``None`` i.e. randomized.
        system_idx: indices of the qubits on which this gate acts on. Defaults to ``None``.
            i.e. list(range(num_acted_qubits)).
        num_acted_param: the number of parameters required for a single operation.
        param_sharing: whether all operations are shared by the same parameter set.
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
        check_legality: whether check the completeness of the matrix if provided. Defaults to ``True``.
        gate_info: information of this gate that will be placed into the gate history or plotted by a Circuit.
            Defaults to ``None``.
        support_batch: whether generator support batch inputs. Defaults to ``False``.
    """

    def __init__(
        self, generator: Callable[[torch.Tensor], torch.Tensor],
        param: Union[torch.Tensor, float, List[float]] = None,
        num_acted_param: int = 1, param_sharing: bool = False,
        system_idx: Union[Iterable[Iterable[int]], Iterable[int], int] = None,
        acted_system_dim: Union[List[int], int] = 2, check_legality: bool = True, gate_info: dict = None,
        support_batch: bool = False
    ):
        # if # of acted system is unknown, generate an example matrix to run Gate.__init__
        ex_matrix = generator(torch.randn(
            [num_acted_param], dtype=get_float_dtype())) if isinstance(acted_system_dim, int) else None

        super().__init__(ex_matrix, system_idx, acted_system_dim, check_legality, gate_info)
        
        intrinsic._theta_generation(self, param, self.system_idx, num_acted_param, param_sharing)
        self.param_sharing = param_sharing
        self.__generator = generator
        self.__support_batch = support_batch
    
    def __call__(self, state: State) -> State:
        return self.forward(state)

    @property
    def matrix(self) -> Union[torch.Tensor, List[torch.Tensor]]:
        theta = self.theta.repeat(len(self.system_idx), 1, 1) if self.param_sharing else self.theta
        
        if self.__support_batch:
            return self.__generator(theta).squeeze().to(device=self.device)
        
        matrices_list = [
            torch.stack([self.__generator(param) for param in theta[param_idx]])
            for param_idx in range(len(self.system_idx))
        ]
        matrices_list = torch.stack(matrices_list)
        return matrices_list.squeeze().to(device=self.device)
    
    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this gate.
        """
        info = super().info
        info.update({
            'param': self.theta,
            'matrix': None if self.theta.shape[1] > 1 else self.matrix,
        })
        return info

    def display_in_circuit(self, ax: matplotlib.axes.Axes, x: float, ) -> float:
        r'''The display function called by circuit instance when plotting.

        Args:
            ax: the ``matplotlib.axes.Axes`` instance
            x: the start horizontal position

        Returns:
            the total width occupied

        Note:
            Users could overload this function for custom display.
        '''
        return _base_param_gate_display(self, ax, x)

    def forward(self, state: State) -> State:
        dim = self.dim
        matrices_list = self.matrix.view([-1, self.theta.shape[-2], dim, dim]).squeeze(-3)
        
        # if self.num_acted_system == 1:
                
        #     # TODO add NKron batch support
        #     state = self._single_qubit_combine_with_threshold6(
        #         state=state, matrices=param_matrices)
        # else:
        #     for param_idx, system_idx in enumerate(self.system_idx):
        #         state._evolve(matrices_list[param_idx], system_idx)
        
        for param_idx, system_idx in enumerate(self.system_idx):
            state._evolve(matrices_list[param_idx], system_idx)
        return state



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the classes for custom quantum channels.
"""

from typing import Iterable, List, Union

import torch

from ...core.intrinsic import _alias
from .base import Channel


class ChoiRepr(Channel):
    r"""A custom channel in Choi representation.

    Args:
        choi_repr: Choi operator of this channel.
        system_idx: Indices of the systems on which this channel acts. Defaults to ``None``.
        acted_system_dim: dimension of systems that this channel acts on. Can be a list of system dimensions
            or an int representing the dimension of all systems. Defaults to be qubit case.

    """

    @_alias({"system_idx": "qubits_idx"})
    def __init__(
        self,
        choi_repr: torch.Tensor,
        system_idx: Union[Iterable[Iterable[int]], Iterable[int], int] = None,
        acted_system_dim: Union[List[int], int] = 2,
    ):
        channel_info = {
            "name": "channel",
            "tex": r'\mathcal{N}',
            "api": "choi_channel",
            "kwargs": {"choi_repr": choi_repr},
        }
        super().__init__(
            "choi", choi_repr, system_idx, acted_system_dim, channel_info=channel_info
        )


class KrausRepr(Channel):
    r"""A custom channel in Kraus representation.

    Args:
        kraus_repr: list of Kraus operators of this channel.
        system_idx: Indices of the systems on which this channel acts. Defaults to ``None``.
        acted_system_dim: dimension of systems that this channel acts on. Can be a list of system dimensions
            or an int representing the dimension of all systems. Defaults to be qubit case.

    """

    @_alias({"system_idx": "qubits_idx"})
    def __init__(
        self,
        kraus_repr: Union[torch.Tensor, List[torch.Tensor]],
        system_idx: Union[Iterable[Iterable[int]], Iterable[int], int] = None,
        acted_system_dim: Union[List[int], int] = 2,
    ):
        channel_info = {
            "name": "channel",
            "tex": r'\mathcal{N}',
            "api": "kraus_channel",
            "kwargs": {"kraus_repr": kraus_repr},
        }
        super().__init__(
            "kraus", kraus_repr, system_idx, acted_system_dim, channel_info=channel_info
        )


class StinespringRepr(Channel):
    r"""A custom channel in Stinespring representation.

    Args:
        stinespring_mat: Stinespring matrix that represents this channel.
        system_idx: Indices of the systems on which this channel acts. Defaults to ``None``.
        acted_system_dim: dimension of systems that this channel acts on. Can be a list of system dimensions
            or an int representing the dimension of all systems. Defaults to be qubit case.

    """

    @_alias({"system_idx": "qubits_idx"})
    def __init__(
        self,
        stinespring_mat: torch.Tensor,
        system_idx: Union[Iterable[Iterable[int]], Iterable[int], int] = None,
        acted_system_dim: Union[List[int], int] = 2,
    ):
        channel_info = {
            "name": "channel",
            "tex": r'\mathcal{N}',
            "api": "stinespring_channel",
            "kwargs": {"stinespring_channel": stinespring_mat},
        }
        super().__init__(
            "stinespring",
            stinespring_mat,
            system_idx,
            acted_system_dim,
            channel_info=channel_info,
        )
